#include "SettingsForm.h"
#include "..\..\common\fast_bit_operation_macros.h"
#include <string.h>

#pragma warning (disable:4800)

using namespace FilterSpace;
using namespace System::Windows::Forms;



/////////////////////////////
//       FILTER            //
/////////////////////////////

bool FilterCore(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	// ITERATOR MAY PREVENT AN UNNEEDED THREAD TO RUN
	bool processing = (f->eliminated_pixels[f->nCoreThreadId] != -1);

	int w = (sw) ? sw : f->width;
	int h = (sh) ? sh : f->height;
	int w_in = f->width;
	int h_in = f->height;

	unsigned char P1, *p = f->dataout, *in = f->datain;
	int           x, y, xx, yy, cnt = 0, s = f->stridein;

	unsigned char count_of_non_zero_neighbours[256];
	for (unsigned int i = 0; i < 256; ++i) // 1-esek megszmllsa
	{
		unsigned int NZ = 0;
		for (unsigned int ui = i; ui; ui &= ui - 1) NZ++;
		count_of_non_zero_neighbours[i] = NZ;
	}

	for (y = 0, yy = sy; y < h; y++, yy++)
	{
		memcpy(p, in + yy * s + sx / 8, sw >> 3);

		if (processing)
		{
			// EROSION -->
			if (f->filtering_iterations_limit < 0)
			{
				for (x = 0, xx = sx; x < w; x++, xx++)
				{
					// 0 = OBJEKTUM-PIXEL, 1 = HTTR-PIXEL
					if (GET_BITPIXEL_OF_ROW(in + yy * s,xx))	continue;

					NEIGHBORHOOD_QUERY(in, xx, yy, w_in, h_in, P1)
					if (count_of_non_zero_neighbours[P1] == 8)	continue;

					SET_WHITEBITPIXEL_OF_ROW(p,x);
					cnt++;
				}
			}
			// DILATION -->
			else
			{
				for (x = 0, xx = sx; x < w; x++, xx++)
				{
					// 0 = OBJEKTUM-PIXEL, 1 = HTTR-PIXEL
					if (GET_BITPIXEL_OF_ROW(in + yy * s,xx) == 0)	continue;

					NEIGHBORHOOD_QUERY(in, xx, yy, w_in, h_in, P1)
					if (count_of_non_zero_neighbours[P1] == 0)		continue;

					SET_BLACKBITPIXEL_OF_ROW(p,x);
					cnt++;
				}
			}
		}
		p += f->strideout;

		// FOR PERFORMANCE MEASUREMENT
		ImageFilterBase::UpdateFilterStatus(f, sw, updater);
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}

	f->eliminated_pixels[f->nCoreThreadId] += cnt;

	// FILTER CODE END -->
	//if (f->processing_ptr && f->processing_size) *(f->processing_ptr) = *(f->processing_size);

	return true;
}



IntPtr ImageFilter::CreateFilterSettings(Byte* fs, bool silent)
{
	int  stride, width, height;
	bool cancelled = false;

	if (!fs) return IntPtr(0);

	xml_node ^xt,  ^xn      = gcnew xml_node( fs );
	filtersettings* f       = XmlToMyStruct( xn, true );
	byte*           indata  = f->datain;
	byte*           outdata = f->dataout;

	if (f && !silent)
	{
		FilterTemplate::SettingsForm^ form = gcnew FilterTemplate::SettingsForm(f);
		form->ShowDialog();
		cancelled = form->FilterCancelled;
		delete form;
	}
	if (cancelled)
	{
		delete f;
		delete xn;
		return IntPtr(0);
	} 

	if (outdata == 0 || indata == outdata)  // FIRST RUN OF CreateFilterSettings
	{
		if (f->width * f->height)
		{
			xn->XmlRemoveChildByName("OutputImage0");

			width  = f->width;
			stride = f->stridein;
			height = f->height;
			outdata = new Byte[stride * height];

			xt = gcnew xml_node( xn, "OutputImage0", "", "" );
			xn->children.Add( xt );
			//                     PinName,  width, height, stride, packed, interleaved, channelcount, pixelsize,       data)
			CreateImageSetting(xt, "Output", width, height, stride,      1,           1,            1,         1,    outdata);
			xt->XmlGetChildByName( xt, "ChannelsBitAllocation" );
			SetChannelsBitAllocation(xt, 0, 1);

			byte binary_color_map[] = { 0, 0, 0, 0, 255, 255, 255, 0 };
			String^ colMap = stringToString(base64((unsigned char*)(binary_color_map), 8));
			xt->XmlSetChildValueByName("ColorMap", colMap, false); 
		}
	}

	xn->XmlSetChildValueByName("FilteringIterationsLimit", f->filtering_iterations_limit);
	xn->XmlSetChildValueByName("Alternating", f->alternating);

	delete f;
	Byte* res = xn->XML_Tree_Marshal();
	delete xn;
	return (IntPtr)(void*)res;
}



filtersettings* ImageFilter::XmlToMyStruct(xml_node^ xn, bool creating)
{
	int packed, interleaved, iTmp;
	xml_node^ xntmp;

	if (! xn->XmlGetChildByName(xntmp, "InputImage0")) return 0;
	filtersettings* filset = new filtersettings;

	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->stridein, packed, interleaved, filset->channelcountin, filset->pixelsizein, filset->datain))
	{
		MessageBox::Show("XmlToMyStruct - 0 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (! xn->XmlGetChildByName(xntmp, "OutputImage0") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 1 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (GetOneImageSetting(xntmp, iTmp, iTmp, filset->strideout, packed, interleaved, filset->channelcountout, filset->pixelsizeout, filset->dataout))
	{
		MessageBox::Show("XmlToMyStruct - 2 - " + creating.ToString());
		delete filset;
		return 0;
	}

	int m = 0;
	if ( xn->XmlGetChildByName(xntmp, "Mask") )	xntmp->XmlGetChildValueByName("Data", m);
	filset->mask = (Byte*)(void*)m;

	filset->processing_size            = ( xn->XmlGetChildValueByName("Processing_Size", m) )  ? (int*)(void*)m : 0;
	filset->processing_ptr             = ( xn->XmlGetChildValueByName("Processing_Ptr", m) )   ? (int*)(void*)m : 0;
	filset->nCoreEnabled               = ( xn->XmlGetChildValueByName("nCore!Enabled", m) )    ? m : 0;
	filset->nCoreKillSignal            = ( xn->XmlGetChildValueByName("nCore!KillSignal", m) ) ? (int*)(void*)m : 0;
	filset->nCoreExecutionSize         = ( xn->XmlGetChildValueByName("nCore!ExecutionSize", m) )      ? m : 2;
	filset->filtering_iterations_limit = ( xn->XmlGetChildValueByName("FilteringIterationsLimit", m) ) ? m : 0; // 0 = skeleton
	filset->alternating	               = ( xn->XmlGetChildValueByName("Alternating", m) ) ? m : 0; // 0,1 = normal, alternating

	return filset;
}


bool ImageFilter::CanLocalize(){ return false; }


IntPtr ImageFilter::GetSupportedFormats()
{
	xml_node ^xn, ^node = gcnew xml_node( nullptr, "", "", "" );

	node->children.Add( gcnew xml_node(node, "InputImageCount", "", "1" ) );
	node->children.Add( gcnew xml_node(node, "OutputImageCount", "", "1" ) );
	xn = gcnew xml_node( node, "InputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Input",     0,      0,      0,      1,           1,            1,         1,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 1);

	xn = gcnew xml_node( node, "OutputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Output",     0,      0,      0,      1,           1,           1,         1,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 1);
	/*
	width,height = 2 --> width, height must be power of 2... = 8 --> multiply of 8
	packed, interleaved = 0, 1 (true, false)
	channelcount = 0 --> any positive number
	channelcount > 0 --> the value itself
	*/

	return (IntPtr)(void*)node->XML_Tree_Marshal();
}






// ***********************************************************************************************
// DO NOT MODIFY OR MOVE TO ELSEWHERE THESE HELPER FUNCTIONS
// THESE ARE HERE BECAUSE OF FilterCore IMPLEMENTED IN THIS FILE!!!

bool ImageFilter::FilterCore(filtersettings* f, int sx, int sy, int sw, int sh)
{
	return false; // (::FilterCore(f, sx, sy, sw, sh, nullptr));
}

bool ImageFilter::ExecuteFilter(Byte* fs)
{ 
	bool b = false;
	if (!fs ) return false;
	xml_node^ xn = gcnew xml_node( fs );
	filtersettings* filset = XmlToMyStruct( xn, false );
	delete xn;
	if (filset)
	{
		filter_core_func f			= (::FilterCore);
		int  plus_one				= 1;
		int	 minus_one				= -1;
		int  eliminated_pixels		= 1;
		int  cpucores				= System::Environment::ProcessorCount;
		bool parallelized			= filset->nCoreEnabled && (cpucores > 1);
		cpucores					= (parallelized) ? cpucores : 1;
		*(filset->processing_size)	= filset->width * filset->height;
		filset->nCoreThreadId		= 0;
		filset->nCoreExecutionSize	= 2; 
		filset->eliminated_pixels	= new int[cpucores];
		memset(filset->eliminated_pixels, 0, sizeof(int) * cpucores); 

		// A FILTER 1X MINDENKPPEN LEFUT. HA BEFEJEZI A FUTST, AKKOR AZ ELIMINLT PIXELEK SZMT 0-RA LLTJA

		memcpy(filset->dataout, filset->datain, filset->stridein * filset->height);
		filset->datain = new byte[filset->stridein * filset->height];

		if (filset->alternating != 0 && filset->filtering_iterations_limit < 0)
			plus_one  = 0;
		else if (filset->alternating != 0)
			minus_one = 0;

		while ((eliminated_pixels > 0) && (filset->filtering_iterations_limit != 0))
		{
			*(filset->processing_ptr) = 0;

			memcpy(filset->datain, filset->dataout, filset->stridein * filset->height);

			b = (parallelized) ? ImageFilterBase::ExecuteFilterBox(filset, f) : f(filset, 0, 0, 0, 0, nullptr);

			if (filset->alternating != 1)
			{
				eliminated_pixels = 0; // VOLT MIT CSINLNI?
				for (int i = 0; i < cpucores; i++) 
				{
					if (filset->eliminated_pixels[i] == 0) filset->eliminated_pixels[i] = -1; // thread group not needed
					if (filset->eliminated_pixels[i]  > 0) 
					{
						eliminated_pixels += filset->eliminated_pixels[i];
						filset->eliminated_pixels[i] = 0;
					}
				}
			}

			filset->filtering_iterations_limit += (filset->filtering_iterations_limit > 0) ? minus_one : plus_one;

			if (filset->alternating) filset->filtering_iterations_limit = -filset->filtering_iterations_limit;
		}

		delete [] filset->datain;
		delete [] filset->eliminated_pixels;
		*(filset->processing_ptr) = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
		delete filset;
	}

	return b;
}

